# -*- coding: utf-8 -*-
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import logging
import time
import datetime
from privacyidea.lib.tokens.HMAC import HmacOtp
from privacyidea.lib.config import get_from_config
from privacyidea.lib.log import log_with
from privacyidea.lib.tokenclass import TokenClass
from privacyidea.lib.tokens.hotptoken import HotpTokenClass
from privacyidea.lib.decorators import check_token_locked
from privacyidea.lib.policy import ACTION, SCOPE, GROUP, Match
from privacyidea.lib.tokens.totptoken import TotpTokenClass
from privacyidea.lib.utils import determine_logged_in_userparams, parse_time_sec_int
from privacyidea.lib import _

log = logging.getLogger(__name__)


class DayPasswordTokenClass(TotpTokenClass):
    previous_otp_offset = 0

    desc_timestep = _('Specify the time step of the DayPassword token. For example: "24h"')

    @log_with(log)
    def __init__(self, db_token):
        """
        Create a new day password token object from a DB Token object

        :param db_token: instance of the orm db object
        :type db_token:  orm object
        """
        TokenClass.__init__(self, db_token)
        self.set_type("daypassword")
        self.hKeyRequired = True

    @staticmethod
    def get_class_type():
        """
        return the token type shortname

        :return: 'daypassword'
        :rtype: string
        """
        return "daypassword"

    @staticmethod
    def get_class_prefix():
        """
        Return the prefix, that is used as a prefix for the serial numbers.

        :return: DayPasswordToken
        """
        return "DYPW"

    @staticmethod
    @log_with(log)
    def get_class_info(key=None, ret='all'):
        """
        returns a subtree of the token definition

        :param key: subsection identifier
        :type key: string
        :param ret: default return value, if nothing is found
        :type ret: user defined
        :return: subsection if key exists or user defined
        :rtype: dict or scalar
        """
        res = {'type': 'daypassword',
               'title': 'Time based Password',
               'description': _('DayPassword: A time-based token with a variable timestep and the possibility'
                                ' to use the OTP more than once.'),
               'user': ['enroll'],
               # This tokentype is enrollable in the UI for...
               'ui_enroll': [SCOPE.ADMIN, SCOPE.USER],
               'policy': {
                   SCOPE.USER: {
                       'timestep': {'type': 'str',
                                    'desc': DayPasswordTokenClass.desc_timestep},
                       'hashlib': {'type': 'str',
                                   'value': ["sha1",
                                             "sha256",
                                             "sha512"],
                                   'desc': DayPasswordTokenClass.desc_hash_func},
                       'otplen': {'type': 'int',
                                  'value': [6, 8],
                                  'desc': DayPasswordTokenClass.desc_otp_len},
                       'force_server_generate': {'type': 'bool',
                                                 'desc': DayPasswordTokenClass.desc_key_gen}
                   },
                   SCOPE.ADMIN: {
                       'timestep': {'type': 'str',
                                    'desc': DayPasswordTokenClass.desc_timestep},
                       'hashlib': {'type': 'str',
                                   'value': ["sha1",
                                             "sha256",
                                             "sha512"],
                                   'desc': DayPasswordTokenClass.desc_hash_func},
                       'otplen': {'type': 'int',
                                  'value': [6, 8],
                                  'desc': DayPasswordTokenClass.desc_otp_len},
                       'force_server_generate': {'type': 'bool',
                                                 'desc': DayPasswordTokenClass.desc_key_gen}
                   },
                   SCOPE.ENROLL: {
                       ACTION.FORCE_APP_PIN: {
                           'type': 'bool',
                           'desc': _('Enforce setting an app pin for the privacyIDEA '
                                     'Authenticator App')
                       },
                       ACTION.MAXTOKENUSER: {
                           'type': 'int',
                           'desc': _("The user may only have this maximum "
                                     "number of daypassword tokens assigned."),
                           'group': GROUP.TOKEN
                       },
                       ACTION.MAXACTIVETOKENUSER: {
                           'type': 'int',
                           'desc': _("The user may only have this maximum number "
                                     "of active daypassword tokens assigned."),
                           'group': GROUP.TOKEN
                       }

                   }
               },
               }
        if key:
            ret = res.get(key, {})
        else:
            if ret == 'all':
                ret = res

        return ret

    @log_with(log)
    def update(self, param, reset_failcount=True):
        """
        This is called during initialization of the token
        to add additional attributes to the token object.

        :param param: dict of initialization parameters
        :type param: dict

        :return: nothing
        """
        HotpTokenClass.update(self, param, reset_failcount=reset_failcount)

        timeStep = param.get("timeStep", self.timestep)
        # we support various hashlib methods, but only on create
        # which is effectively set in the update
        hashlibStr = param.get("hashlib", self.hashlib)

        self.add_tokeninfo("timeStep", timeStep)
        self.add_tokeninfo("hashlib", hashlibStr)

    @property
    def timestep(self):
        timeStepping = parse_time_sec_int(self.get_tokeninfo("timeStep") or
                                          (get_from_config("daypassword.timeStep")) or "1d")

        return timeStepping

    @property
    def hashlib(self):
        hashlibStr = self.get_tokeninfo("hashlib") or \
                     get_from_config("daypassword.hashlib", 'sha1')
        return hashlibStr

    @log_with(log)
    def check_otp_exist(self, otp, options=None, symetric=False,
                        inc_counter=True):
        """
        checks if the given password value is/are values of this very token at all.
        This is used to autoassign and to determine the serial number of
        a token.

        :param otp: the to be verified otp value
        :type otp: string
        :return: counter or -1 if otp does not exist
        :rtype:  int
        """
        options = options or {}
        res = self.check_otp(otp, options=options)

        if inc_counter and res >= 0:
            # As usually the counter is increased in lib.token.checkUserPass,
            # we need to do this manually here:
            self.inc_otp_counter(res)
        return res

    @check_token_locked
    def check_otp(self, anOtpVal, counter=None, options=None):
        """
        validate the token passwort against a given passwordvalue

        :param anOtpVal: the to be verified passwordvalue
        :type anOtpVal:  string
        :param counter: the counter state, that should be verified. For DayPasswordToken
        this is the unix system time (seconds) divided by 30/60
        :type counter: int
        :param options: the dict, which could contain token specific info
        :type options: dict
        :return: the counter or -1
        :rtype: int
        """
        otplen = int(self.token.otplen)
        options = options or {}
        secretHOtp = self.token.get_otpkey()
        oCount = self.get_otp_count()
        inow = int(time.time())

        initTime = int(options.get('initTime', -1))
        if initTime != -1:
            server_time = int(initTime)
        else:
            server_time = time.time()

        # If we have a counter from the parameter list
        if not counter:
            # No counter, so we take the current token_time
            counter = self._time2counter(server_time,
                                         timeStepping=self.timestep)

        hmac2Otp = HmacOtp(secretHOtp,
                           counter,
                           otplen,
                           self.get_hashlib(self.hashlib))
        res = hmac2Otp.checkOtp(anOtpVal,
                                int(1),
                                symetric=False)

        if res != -1:
            # on success, we have to save the last attempt
            self.set_otp_count(res)

            # here we calculate the new drift/shift between the server time
            # and the tokentime
            tokentime = self._counter2time(res, self.timestep)
            tokenDt = datetime.datetime.fromtimestamp(tokentime / 1.0)

            nowDt = datetime.datetime.fromtimestamp(inow / 1.0)

            lastauth = self._counter2time(oCount, self.timestep)
            lastauthDt = datetime.datetime.fromtimestamp(lastauth / 1.0)

            log.debug("last auth : {0!r}".format(lastauthDt))
            log.debug("tokentime : {0!r}".format(tokenDt))
            log.debug("now       : {0!r}".format(nowDt))
            log.debug("delta     : {0!r}".format((tokentime - inow)))

        return res

    def get_otp(self, current_time=None, do_truncation=True,
                time_seconds=None, challenge=None):
        """
        get the next OTP value

        :param current_time: the current time, for which the OTP value
        should be calculated for.
        :type current_time: datetime object
        :param time_seconds: the current time, for which the OTP value
        should be calculated for (date +%s)
        :type: time_seconds: int, unix system time seconds
        :return: next otp value, and PIN, if possible
        :rtype: tuple
        """
        otplen = int(self.token.otplen)
        secretHOtp = self.token.get_otpkey()

        hmac2Otp = HmacOtp(secretHOtp,
                           self.get_otp_count(),
                           otplen,
                           self.get_hashlib(self.hashlib))

        if time_seconds is None:
            time_seconds = time.time()
        if current_time:
            time_seconds = self._time2float(current_time)

        counter = int(time_seconds / self.timestep)
        otpval = hmac2Otp.generate(counter=counter,
                                   inc_counter=False,
                                   do_truncation=do_truncation,
                                   challenge=challenge)

        pin = self.token.get_pin()
        combined = "{0!s}{1!s}".format(otpval, pin)
        if get_from_config("PrependPin") == "True":
            combined = "{0!s}{1!s}".format(pin, otpval)

        return 1, pin, otpval, combined

    @log_with(log)
    def get_multi_otp(self, count=0, epoch_start=0, epoch_end=0,
                      curTime=None, timestamp=None):
        """
        return a dictionary of multiple future OTP values
        of the HOTP/HMAC token

        :param count: how many otp values should be returned
        :type count: int
        :param epoch_start: not implemented
        :param epoch_end: not implemented
        :param curTime: Simulate the servertime
        :type curTime: datetime.datetime
        :param timestamp: Simulate the servertime
        :type timestamp: epoch time
        :return: tuple of status: boolean, error: text and the OTP dictionary

        """
        otp_dict = {"type": "DayPasswordToken", "otp": {}}
        ret = False
        error = "No count specified"

        otplen = int(self.token.otplen)
        secretHOtp = self.token.get_otpkey()

        hmac2Otp = HmacOtp(secretHOtp, self.get_otp_count(),
                           otplen, self.get_hashlib(self.hashlib))

        if curTime:
            # datetime object provided for simulation
            tCounter = self._time2float(curTime)
        elif timestamp:
            # epoch time provided for simulation
            tCounter = int(timestamp)
        else:
            # use the current server time
            tCounter = int(time.time())

        counter = int(tCounter / self.timestep)

        if count > 0:
            error = "OK"
            for i in range(0, count):
                otpval = hmac2Otp.generate(counter=counter + i,
                                           inc_counter=False)
                timeCounter = (counter + i) * self.timestep

                val_time = datetime.datetime. \
                    fromtimestamp(timeCounter).strftime("%Y-%m-%d %H:%M:%S")
                otp_dict["otp"][counter + i] = {'otpval': otpval,
                                                'time': val_time}
            ret = True

        return ret, error, otp_dict

    @staticmethod
    def get_setting_type(key):
        settings = {"daypassword.hashlib": "public",
                    "daypassword.timeStep": "public"}
        return settings.get(key, "")

    @classmethod
    def get_default_settings(cls, g, params):
        """
        This method returns a dictionary with default settings for token
        enrollment.
        These default settings are defined in SCOPE.USER or SCOPE.ADMIN and are
        daypassword_hashlib, daypassword_timestep and daypassword_otplen.
        If these are set, the user or admin will only be able to enroll tokens
        with these values.

        The returned dictionary is added to the parameters of the API call.

        :param g: context object, see documentation of ``Match``
        :param params: The call parameters
        :type params: dict
        :return: default parameters
        """
        ret = {}
        if not g.logged_in_user:
            return ret
        (role, username, userrealm, adminuser, adminrealm) = determine_logged_in_userparams(g.logged_in_user,
                                                                                            params)
        hashlib_pol = Match.generic(g, scope=role,
                                    action="daypassword_hashlib",
                                    user=username,
                                    realm=userrealm,
                                    adminuser=adminuser,
                                    adminrealm=adminrealm).action_values(unique=True)
        if hashlib_pol:
            ret["hashlib"] = list(hashlib_pol)[0]

        timestep_pol = Match.generic(g, scope=role,
                                     action="daypassword_timestep",
                                     user=username,
                                     realm=userrealm,
                                     adminuser=adminuser,
                                     adminrealm=adminrealm).action_values(unique=True)
        if timestep_pol:
            ret["timeStep"] = list(timestep_pol)[0]

        otplen_pol = Match.generic(g, scope=role,
                                   action="daypassword_otplen",
                                   user=username,
                                   realm=userrealm,
                                   adminuser=adminuser,
                                   adminrealm=adminrealm).action_values(unique=True)
        if otplen_pol:
            ret["otplen"] = list(otplen_pol)[0]

        return ret

    @staticmethod
    def get_import_csv(l):
        """
        Read the list from a csv file and return a dictionary, that can be used
        to do a token_init.

        :param l: The list of the line of a csv file
        :type l: list
        :return: A dictionary of init params
        """
        params = TokenClass.get_import_csv(l)
        # timeStep
        if len(l) >= 5:
            params["timeStep"] = int(l[4].strip())
        else:
            params["timeStep"] = 30

        return params



"""v3.9: Create sequences needed for SQLAlchemy 1.4

E.g. mariadb now needs sequences.

Revision ID: 5cb310101a1f
Revises: 4a0aec37e7cf
Create Date: 2023-09-08 15:59:01.374626

"""

# revision identifiers, used by Alembic.
revision = '5cb310101a1f'
down_revision = '4a0aec37e7cf'

from alembic import op, context
from sqlalchemy.schema import Sequence, CreateSequence, DropSequence
from sqlalchemy.orm import sessionmaker
from sqlalchemy.sql.expression import func
from sqlalchemy.exc import OperationalError, ProgrammingError
from privacyidea.models import db


Session = sessionmaker()


def upgrade():
    migration_context = context.get_context()
    if migration_context.dialect.supports_sequences:
        if migration_context.dialect.name in ['mariadb', 'mysql']:
            # setting "increment" to 0 works like auto-increment but also supports replication
            # See <https://mariadb.com/kb/en/sequence-overview/#replication>
            inc_value = 0
        else:
            inc_value = 1
        bind = op.get_bind()
        # We only need a read session, so we do not need a commit
        session = Session(bind=bind)
        # Loop over all tables defined in the current models.py
        for tbl in db.metadata.sorted_tables:
            # we act only on tables with an "id" column
            if 'id' in tbl.c:
                seq = tbl.c['id'].default
                if isinstance(seq, Sequence):
                    seq_name = seq.name
                else:
                    # do we have tables with "id" columns which isn't a sequence?
                    print(f"Table {tbl.name} has an 'id' column which isn't a sequence!")
                    continue

                # Create the sequence with the correct next_id!
                current_id = session.query(func.max(tbl.c.id)).one()[0] or 0
                print(f"CurrentID in Table {tbl.name}: {current_id}")
                try:
                    seq = Sequence(seq_name, start=(current_id + 1), increment=inc_value)
                    print(f" +++ Creating Sequence: {seq_name}")
                    op.execute(CreateSequence(seq, if_not_exists=True))
                except OperationalError as exx:
                    if exx.orig.args[0] == 1050 or "already exists" in exx.orig.args[1]:
                        pass
                    else:
                        print(exx)
                except ProgrammingError as exx:
                    if "already exists" in exx.orig.args[0]:
                        pass
                    else:
                        print(exx)
                except Exception as exx:
                    print(exx)


def downgrade():

    migration_context = context.get_context()
    if migration_context.dialect.supports_sequences and migration_context.dialect.name in ['mysql', 'mariadb']:
        # don't remove sequences unless the db is MariaDB
        for tbl in db.metadata.sorted_tables:
            if 'id' in tbl.c:
                seq = tbl.c.id.default
                if isinstance(seq, Sequence):
                    try:
                        op.execute(DropSequence(seq))
                    except Exception as exx:
                        print(exx)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  2022-02-16 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Init
#
from __future__ import print_function
__doc__ = """You can use this script to migrate tokens from one privacyIDEA instance to another privacyIDEA instance.
The token data is copied on a database level thus preserving the hashed PIN and the token seeds.
Note: The same encryption keys need to be used.

You need to provide the two pi.cfg files of the two instances. 
Note: Using the same pi.cfg file, you could copy tokens within one instance.

In the sections MIGRATE->user you can specify, which users you want to find, so that the
script does not need to iterate through all users.
The pattern defines a search pattern on the username, that needs to match.
Rules for regular expression apply.
You can then define the new username in "replace".
The "attributes" define, which user attributes should be copied. Then the user is created on the new instance.

All tokens of this found user are migrated. However, you can use the section MIGRATE->serial to change the
serial number of the migrated tokens.

The section ASSIGNMENTS defines, in which resolver and realm the new user shall be created.

"""
from sqlalchemy.schema import Sequence
import sys
import json
import getopt

import re
from privacyidea.models import TokenInfo, MethodsMixin
from privacyidea.app import create_app
from privacyidea.lib.user import get_user_list, create_user, User
from privacyidea.lib.token import get_tokens, assign_token
from flask_sqlalchemy import SQLAlchemy

db = SQLAlchemy()

EXAMPLE_CONFIG_FILE = """{
    "SQL": {
        "PRIVACYIDEA_FROM": "/etc/privacyidea/pi.cfg",
        "PRIVACYIDEA_TO": "/etc/newprivacyidea/pi.cfg"
    },
    "MIGRATE": {
        "user": { "find": {"username": "*@example.com", "realm": "RealmA" },
                  "pattern": "^(.*?)@example.com$",
                  "replace": "\\1",
                  "attributes": ['email', 'givenname', 'surname']},
        "serial": { "pattern": "^(.*)$",
                    "replace": "\\_new" }
    },
    "ASSIGNMENTS": {
        "to_realm": "realmC",
        "to_resolver": "resolverC"
    }
}"""


class Token(MethodsMixin, db.Model):
    """
    The "Token" table contains the basic token data.

    It contains data like
     * serial number
     * secret key
     * PINs
     * ...

    The table :py:class:`privacyidea.models.TokenOwner` contains the owner
    information of the specified token.
    The table :py:class:`privacyidea.models.TokenInfo` contains additional information
    that is specific to the tokentype.
    """
    __tablename__ = 'token'
    __table_args__ = {'mysql_row_format': 'DYNAMIC'}
    id = db.Column(db.Integer, Sequence("token_seq"),
                   primary_key=True,
                   nullable=False)
    description = db.Column(db.Unicode(80), default=u'')
    serial = db.Column(db.Unicode(40), default=u'',
                       unique=True,
                       nullable=False,
                       index=True)
    tokentype = db.Column(db.Unicode(30),
                          default=u'HOTP',
                          index=True)
    user_pin = db.Column(db.Unicode(512),
                         default=u'')  # encrypt
    user_pin_iv = db.Column(db.Unicode(32),
                            default=u'')  # encrypt
    so_pin = db.Column(db.Unicode(512),
                       default=u'')  # encrypt
    so_pin_iv = db.Column(db.Unicode(32),
                          default=u'')  # encrypt
    pin_seed = db.Column(db.Unicode(32),
                         default=u'')
    otplen = db.Column(db.Integer(),
                       default=6)
    pin_hash = db.Column(db.Unicode(512),
                         default=u'')  # hashed
    key_enc = db.Column(db.Unicode(1024),
                        default=u'')  # encrypt
    key_iv = db.Column(db.Unicode(32),
                       default=u'')
    maxfail = db.Column(db.Integer(),
                        default=10)
    active = db.Column(db.Boolean(),
                       nullable=False,
                       default=True)
    revoked = db.Column(db.Boolean(),
                        default=False)
    locked = db.Column(db.Boolean(),
                       default=False)
    failcount = db.Column(db.Integer(),
                          default=0)
    count = db.Column(db.Integer(),
                      default=0)
    count_window = db.Column(db.Integer(),
                             default=10)
    sync_window = db.Column(db.Integer(),
                            default=1000)
    rollout_state = db.Column(db.Unicode(10),
                              default=u'')

    def __init__(self, serial, tokentype=u"", otplen=6,
                 key_enc=None, key_iv=None, description=None,
                 pin_seed=None, count=0, failcount=0, maxfail=10,
                 active=True, locked=False, revoked=False,
                 count_window=10, pin_hash=None, sync_window=1000,
                 rollout_state="",
                 **kwargs):
        super(Token, self).__init__(**kwargs)
        self.serial = u'' + serial
        self.tokentype = tokentype
        self.count = count
        self.failcount = failcount
        self.maxfail = maxfail
        self.active = active
        self.revoked = revoked
        self.locked = locked
        self.count_window = count_window
        self.otplen = otplen
        self.pin_seed = pin_seed
        self.key_enc = key_enc
        self.key_iv = key_iv
        self.description = description
        self.pin_hash = pin_hash,
        self.sync_window = sync_window
        self.rollout_state = rollout_state


class Config(object):

    def __init__(self, config_file):
        with open(config_file, "r") as f:
            contents = f.read()
        config = json.loads(contents)
        self.ASSIGNMENTS = config.get("ASSIGNMENTS")
        self.PRIVACYIDEA_FROM = config.get("SQL").get("PRIVACYIDEA_FROM")
        self.PRIVACYIDEA_TO = config.get("SQL").get("PRIVACYIDEA_TO")
        self.INSERT_CHUNK_SIZE = config.get("SQL").get("INSERT_CHUNK_SIZE")
        self.MIGRATE = config.get("MIGRATE")
        self.MIGRATE_USER = self.MIGRATE.get("user")
        self.MIGRATE_USER_FIND = self.MIGRATE_USER.get("find")
        self.MIGRATE_USER_PATTERN = self.MIGRATE_USER.get("pattern")
        self.MIGRATE_USER_REPLACE = self.MIGRATE_USER.get("replace")
        self.TO_RESOLVER = self.ASSIGNMENTS.get("to_resolver")
        self.TO_REALM = self.ASSIGNMENTS.get("to_realm")
        self.MIGRATE_ATTRIBUTES = self.MIGRATE.get("user").get("attributes", [])
        self.MIGRATE_SERIAL_PATTERN = self.MIGRATE.get("serial", {}).get("pattern")
        self.MIGRATE_SERIAL_REPLACE = self.MIGRATE.get("serial", {}).get("replace")


def dict_without_keys(d, keys):
    new_d = d.copy()
    for key in keys:
        if key in d:
            new_d.pop(key)
    return new_d


def token_to_dict(token):
    """
    Store the database columns of the token into a dict.
    Also store the tokeninfo into a list of dicts.

    :param token: The database token object
    :return: a dict, containing the token and the tokeninfo
    """
    token_dict = {}
    columns = token.__table__.c
    for column in columns:
        value = getattr(token, column.key)
        if column.key not in ('id'):
            token_dict[column.key] = value
    # Now add the tokeninfo
    info_list = []
    for ti in token.info_list:
        tokeninfo = {"Description": ti.Description,
                     "Key": ti.Key,
                     "Type": ti.Type,
                     "Value": ti.Value}
        info_list.append(tokeninfo)
    token_dict["info_list"] = info_list
    return token_dict


def create_token_from_dict(serialized_token, info_list):
    """
    
    :param serialized_token: dict containing all token objects 
    :return: database ID of the token
    """
    # create database object directly, since we have the encrypted data
    r = Token(**serialized_token).save()
    for ti in info_list:
        ti["token_id"] = r
        TokenInfo(**ti).save()
    return r


def migrate(config_obj):

    from_app = create_app(config_name="production",
                          config_file=config_obj.PRIVACYIDEA_FROM,
                          silent=True)

    to_app = create_app(config_name="production",
                        config_file=config_obj.PRIVACYIDEA_TO,
                        silent=True)

    new_users = []
    new_tokens = []

    with from_app.app_context():
        # find all the users
        userlist = get_user_list(param=config_obj.MIGRATE_USER_FIND)
        for user in userlist:
            if re.match(config_obj.MIGRATE_USER_PATTERN, user.get("username")):
                new_username = re.sub(config_obj.MIGRATE_USER_PATTERN, config_obj.MIGRATE_USER_REPLACE, user.get("username"))
                new_user = {"username": new_username,
                            "tokenlist": []}
                for attr in config_obj.MIGRATE_ATTRIBUTES:
                    new_user[attr] = user.get(attr)

                tokens = get_tokens(user=User(user.get("username"), realm=config_obj.MIGRATE_USER_FIND.get("realm")))
                for token in tokens:
                    new_tokens.append(token_to_dict(token.token))
                    new_user["tokenlist"].append(token.token.serial)
                new_users.append(new_user)

    with to_app.app_context():
        # create the new tokens
        for tok in new_tokens:
            if config_obj.MIGRATE_SERIAL_PATTERN:
                tok["serial"] = re.sub(config_obj.MIGRATE_SERIAL_PATTERN,
                                       config_obj.MIGRATE_SERIAL_REPLACE,
                                       tok["serial"])
            info_list = tok.get("info_list")
            del (tok["info_list"])
            toks = get_tokens(serial=tok.get("serial"))
            if len(toks) > 0:
                print("New token {0!s} aleady exists.".format(tok.get("serial")))
            else:
                create_token_from_dict(tok, info_list)

        # create the new users
        for user in new_users:
            tokenlist = user.get("tokenlist")
            del(user["tokenlist"])

            ul = get_user_list({"username": user.get("username"),
                                "realm": config_obj.TO_REALM,
                                "resolver": config_obj.TO_RESOLVER})
            if not ul:
                uid = create_user(config_obj.TO_RESOLVER, user)
                print("Created user {0!s}".format(uid))
            else:
                print("User already exists!")
            user_obj = User(login=user.get("username"),
                            realm=config_obj.TO_REALM,
                            resolver=config_obj.TO_RESOLVER)

            # Assign token
            for serial in tokenlist:
                serial = re.sub(config_obj.MIGRATE_SERIAL_PATTERN,
                               config_obj.MIGRATE_SERIAL_REPLACE,
                               serial)
                print("Assigning token {0!s} to user {1!s}".format(serial, user_obj))
                try:
                    assign_token(serial, user_obj)
                except Exception:
                    print("Error assigning token - probably the token is already assigned.")


def usage():
    print("""
migrate-token.py --generate-example-config [--config <config file>]

    --generate-example-config, -g   Output an example config file. 
                                    This is a JSON file, that needs to be passed
                                    to this command.

    --config, -c <file>             The config file, that contains the complete
                                    configuration.

{0!s}
""".format(__doc__))


def main():
    try:
        opts, args = getopt.getopt(sys.argv[1:], "hgc:", ["help", "generate-example-config", "config="])
    except getopt.GetoptError as e:
        print(str(e))
        sys.exit(1)

    config_file = None
    generate_config = False

    for o, a in opts:
        if o in ("-g", "--generate-example-config"):
            generate_config = True
            print(EXAMPLE_CONFIG_FILE)
        elif o in ("-c", "--config"):
            config_file = a
        elif o in ("-h", "--help"):
            usage()
            sys.exit(2)

    if config_file:
        config_obj = Config(config_file)
        migrate(config_obj)
        sys.exit(0)

    else:
        if not generate_config:
            usage()
            sys.exit(1)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: caservice.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0f\x63\x61service.proto\"\'\n\x06Status\x12\x0c\n\x04\x63ode\x18\x01 \x01(\x05\x12\x0f\n\x07message\x18\x02 \x01(\t\";\n\x10SetOptionRequest\x12\x12\n\noptionName\x18\x01 \x01(\t\x12\x13\n\x0boptionValue\x18\x02 \x01(\t\")\n\x0eSetOptionReply\x12\x17\n\x06status\x18\x01 \x01(\x0b\x32\x07.Status\"\x13\n\x11GetOptionsRequest\"\x8a\x01\n\x0fGetOptionsReply\x12.\n\x07options\x18\x01 \x03(\x0b\x32\x1d.GetOptionsReply.OptionsEntry\x12\x17\n\x06status\x18\x02 \x01(\x0b\x32\x07.Status\x1a.\n\x0cOptionsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"E\n\x10SubmitCSRRequest\x12\x0b\n\x03\x63sr\x18\x01 \x01(\t\x12\x14\n\x0ctemplateName\x18\x02 \x01(\t\x12\x0e\n\x06\x63\x61Name\x18\x03 \x01(\t\"m\n\x0eSubmitCSRReply\x12\x13\n\x0b\x64isposition\x18\x01 \x01(\x05\x12\x1a\n\x12\x64ispositionMessage\x18\x02 \x01(\t\x12\x11\n\trequestId\x18\x03 \x01(\x05\x12\x17\n\x06status\x18\x05 \x01(\x0b\x32\x07.Status\"\x0f\n\rGetCAsRequest\"7\n\x0bGetCAsReply\x12\x0f\n\x07\x63\x61Names\x18\x01 \x03(\t\x12\x17\n\x06status\x18\x02 \x01(\x0b\x32\x07.Status\"3\n\x15GetCertificateRequest\x12\x0e\n\x06\x63\x61Name\x18\x01 \x01(\t\x12\n\n\x02id\x18\x02 \x01(\x05\"<\n\x13GetCertificateReply\x12\x0c\n\x04\x63\x65rt\x18\x01 \x01(\t\x12\x17\n\x06status\x18\x02 \x01(\x0b\x32\x07.Status\"%\n\x13GetTemplatesRequest\x12\x0e\n\x06\x63\x61Name\x18\x01 \x01(\t\"C\n\x11GetTemplatesReply\x12\x15\n\rtemplateNames\x18\x01 \x03(\t\x12\x17\n\x06status\x18\x02 \x01(\x0b\x32\x07.Status\"b\n\x13GetCSRStatusRequest\x12\x0e\n\x06\x63\x61Name\x18\x01 \x01(\t\x12\x17\n\rcertRequestId\x18\x02 \x01(\x05H\x00\x12\x14\n\ncertSerial\x18\x03 \x01(\tH\x00\x42\x0c\n\nIDorSerial\"]\n\x11GetCSRStatusReply\x12\x13\n\x0b\x64isposition\x18\x01 \x01(\x05\x12\x1a\n\x12\x64ispositionMessage\x18\x02 \x01(\t\x12\x17\n\x06status\x18\x03 \x01(\x0b\x32\x07.Status*\x19\n\x04\x43ode\x12\x06\n\x02OK\x10\x00\x12\t\n\x05\x45RROR\x10\x01\x32\xfd\x02\n\tCAService\x12/\n\tSubmitCSR\x12\x11.SubmitCSRRequest\x1a\x0f.SubmitCSRReply\x12&\n\x06GetCAs\x12\x0e.GetCAsRequest\x1a\x0c.GetCAsReply\x12>\n\x0eGetCertificate\x12\x16.GetCertificateRequest\x1a\x14.GetCertificateReply\x12\x38\n\x0cGetTemplates\x12\x14.GetTemplatesRequest\x1a\x12.GetTemplatesReply\x12\x38\n\x0cGetCSRStatus\x12\x14.GetCSRStatusRequest\x1a\x12.GetCSRStatusReply\x12/\n\tSetOption\x12\x11.SetOptionRequest\x1a\x0f.SetOptionReply\x12\x32\n\nGetOptions\x12\x12.GetOptionsRequest\x1a\x10.GetOptionsReplyB\x07\xaa\x02\x04Grpcb\x06proto3')

_CODE = DESCRIPTOR.enum_types_by_name['Code']
Code = enum_type_wrapper.EnumTypeWrapper(_CODE)
OK = 0
ERROR = 1


_STATUS = DESCRIPTOR.message_types_by_name['Status']
_SETOPTIONREQUEST = DESCRIPTOR.message_types_by_name['SetOptionRequest']
_SETOPTIONREPLY = DESCRIPTOR.message_types_by_name['SetOptionReply']
_GETOPTIONSREQUEST = DESCRIPTOR.message_types_by_name['GetOptionsRequest']
_GETOPTIONSREPLY = DESCRIPTOR.message_types_by_name['GetOptionsReply']
_GETOPTIONSREPLY_OPTIONSENTRY = _GETOPTIONSREPLY.nested_types_by_name['OptionsEntry']
_SUBMITCSRREQUEST = DESCRIPTOR.message_types_by_name['SubmitCSRRequest']
_SUBMITCSRREPLY = DESCRIPTOR.message_types_by_name['SubmitCSRReply']
_GETCASREQUEST = DESCRIPTOR.message_types_by_name['GetCAsRequest']
_GETCASREPLY = DESCRIPTOR.message_types_by_name['GetCAsReply']
_GETCERTIFICATEREQUEST = DESCRIPTOR.message_types_by_name['GetCertificateRequest']
_GETCERTIFICATEREPLY = DESCRIPTOR.message_types_by_name['GetCertificateReply']
_GETTEMPLATESREQUEST = DESCRIPTOR.message_types_by_name['GetTemplatesRequest']
_GETTEMPLATESREPLY = DESCRIPTOR.message_types_by_name['GetTemplatesReply']
_GETCSRSTATUSREQUEST = DESCRIPTOR.message_types_by_name['GetCSRStatusRequest']
_GETCSRSTATUSREPLY = DESCRIPTOR.message_types_by_name['GetCSRStatusReply']
Status = _reflection.GeneratedProtocolMessageType('Status', (_message.Message,), {
  'DESCRIPTOR' : _STATUS,
  '__module__' : 'caservice_pb2'
  # @@protoc_insertion_point(class_scope:Status)
  })
_sym_db.RegisterMessage(Status)

SetOptionRequest = _reflection.GeneratedProtocolMessageType('SetOptionRequest', (_message.Message,), {
  'DESCRIPTOR' : _SETOPTIONREQUEST,
  '__module__' : 'caservice_pb2'
  # @@protoc_insertion_point(class_scope:SetOptionRequest)
  })
_sym_db.RegisterMessage(SetOptionRequest)

SetOptionReply = _reflection.GeneratedProtocolMessageType('SetOptionReply', (_message.Message,), {
  'DESCRIPTOR' : _SETOPTIONREPLY,
  '__module__' : 'caservice_pb2'
  # @@protoc_insertion_point(class_scope:SetOptionReply)
  })
_sym_db.RegisterMessage(SetOptionReply)

GetOptionsRequest = _reflection.GeneratedProtocolMessageType('GetOptionsRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETOPTIONSREQUEST,
  '__module__' : 'caservice_pb2'
  # @@protoc_insertion_point(class_scope:GetOptionsRequest)
  })
_sym_db.RegisterMessage(GetOptionsRequest)

GetOptionsReply = _reflection.GeneratedProtocolMessageType('GetOptionsReply', (_message.Message,), {

  'OptionsEntry' : _reflection.GeneratedProtocolMessageType('OptionsEntry', (_message.Message,), {
    'DESCRIPTOR' : _GETOPTIONSREPLY_OPTIONSENTRY,
    '__module__' : 'caservice_pb2'
    # @@protoc_insertion_point(class_scope:GetOptionsReply.OptionsEntry)
    })
  ,
  'DESCRIPTOR' : _GETOPTIONSREPLY,
  '__module__' : 'caservice_pb2'
  # @@protoc_insertion_point(class_scope:GetOptionsReply)
  })
_sym_db.RegisterMessage(GetOptionsReply)
_sym_db.RegisterMessage(GetOptionsReply.OptionsEntry)

SubmitCSRRequest = _reflection.GeneratedProtocolMessageType('SubmitCSRRequest', (_message.Message,), {
  'DESCRIPTOR' : _SUBMITCSRREQUEST,
  '__module__' : 'caservice_pb2'
  # @@protoc_insertion_point(class_scope:SubmitCSRRequest)
  })
_sym_db.RegisterMessage(SubmitCSRRequest)

SubmitCSRReply = _reflection.GeneratedProtocolMessageType('SubmitCSRReply', (_message.Message,), {
  'DESCRIPTOR' : _SUBMITCSRREPLY,
  '__module__' : 'caservice_pb2'
  # @@protoc_insertion_point(class_scope:SubmitCSRReply)
  })
_sym_db.RegisterMessage(SubmitCSRReply)

GetCAsRequest = _reflection.GeneratedProtocolMessageType('GetCAsRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETCASREQUEST,
  '__module__' : 'caservice_pb2'
  # @@protoc_insertion_point(class_scope:GetCAsRequest)
  })
_sym_db.RegisterMessage(GetCAsRequest)

GetCAsReply = _reflection.GeneratedProtocolMessageType('GetCAsReply', (_message.Message,), {
  'DESCRIPTOR' : _GETCASREPLY,
  '__module__' : 'caservice_pb2'
  # @@protoc_insertion_point(class_scope:GetCAsReply)
  })
_sym_db.RegisterMessage(GetCAsReply)

GetCertificateRequest = _reflection.GeneratedProtocolMessageType('GetCertificateRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETCERTIFICATEREQUEST,
  '__module__' : 'caservice_pb2'
  # @@protoc_insertion_point(class_scope:GetCertificateRequest)
  })
_sym_db.RegisterMessage(GetCertificateRequest)

GetCertificateReply = _reflection.GeneratedProtocolMessageType('GetCertificateReply', (_message.Message,), {
  'DESCRIPTOR' : _GETCERTIFICATEREPLY,
  '__module__' : 'caservice_pb2'
  # @@protoc_insertion_point(class_scope:GetCertificateReply)
  })
_sym_db.RegisterMessage(GetCertificateReply)

GetTemplatesRequest = _reflection.GeneratedProtocolMessageType('GetTemplatesRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETTEMPLATESREQUEST,
  '__module__' : 'caservice_pb2'
  # @@protoc_insertion_point(class_scope:GetTemplatesRequest)
  })
_sym_db.RegisterMessage(GetTemplatesRequest)

GetTemplatesReply = _reflection.GeneratedProtocolMessageType('GetTemplatesReply', (_message.Message,), {
  'DESCRIPTOR' : _GETTEMPLATESREPLY,
  '__module__' : 'caservice_pb2'
  # @@protoc_insertion_point(class_scope:GetTemplatesReply)
  })
_sym_db.RegisterMessage(GetTemplatesReply)

GetCSRStatusRequest = _reflection.GeneratedProtocolMessageType('GetCSRStatusRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETCSRSTATUSREQUEST,
  '__module__' : 'caservice_pb2'
  # @@protoc_insertion_point(class_scope:GetCSRStatusRequest)
  })
_sym_db.RegisterMessage(GetCSRStatusRequest)

GetCSRStatusReply = _reflection.GeneratedProtocolMessageType('GetCSRStatusReply', (_message.Message,), {
  'DESCRIPTOR' : _GETCSRSTATUSREPLY,
  '__module__' : 'caservice_pb2'
  # @@protoc_insertion_point(class_scope:GetCSRStatusReply)
  })
_sym_db.RegisterMessage(GetCSRStatusReply)

_CASERVICE = DESCRIPTOR.services_by_name['CAService']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\252\002\004Grpc'
  _GETOPTIONSREPLY_OPTIONSENTRY._options = None
  _GETOPTIONSREPLY_OPTIONSENTRY._serialized_options = b'8\001'
  _CODE._serialized_start=1000
  _CODE._serialized_end=1025
  _STATUS._serialized_start=19
  _STATUS._serialized_end=58
  _SETOPTIONREQUEST._serialized_start=60
  _SETOPTIONREQUEST._serialized_end=119
  _SETOPTIONREPLY._serialized_start=121
  _SETOPTIONREPLY._serialized_end=162
  _GETOPTIONSREQUEST._serialized_start=164
  _GETOPTIONSREQUEST._serialized_end=183
  _GETOPTIONSREPLY._serialized_start=186
  _GETOPTIONSREPLY._serialized_end=324
  _GETOPTIONSREPLY_OPTIONSENTRY._serialized_start=278
  _GETOPTIONSREPLY_OPTIONSENTRY._serialized_end=324
  _SUBMITCSRREQUEST._serialized_start=326
  _SUBMITCSRREQUEST._serialized_end=395
  _SUBMITCSRREPLY._serialized_start=397
  _SUBMITCSRREPLY._serialized_end=506
  _GETCASREQUEST._serialized_start=508
  _GETCASREQUEST._serialized_end=523
  _GETCASREPLY._serialized_start=525
  _GETCASREPLY._serialized_end=580
  _GETCERTIFICATEREQUEST._serialized_start=582
  _GETCERTIFICATEREQUEST._serialized_end=633
  _GETCERTIFICATEREPLY._serialized_start=635
  _GETCERTIFICATEREPLY._serialized_end=695
  _GETTEMPLATESREQUEST._serialized_start=697
  _GETTEMPLATESREQUEST._serialized_end=734
  _GETTEMPLATESREPLY._serialized_start=736
  _GETTEMPLATESREPLY._serialized_end=803
  _GETCSRSTATUSREQUEST._serialized_start=805
  _GETCSRSTATUSREQUEST._serialized_end=903
  _GETCSRSTATUSREPLY._serialized_start=905
  _GETCSRSTATUSREPLY._serialized_end=998
  _CASERVICE._serialized_start=1028
  _CASERVICE._serialized_end=1409
# @@protoc_insertion_point(module_scope)
